from collections import OrderedDict
from enum import Enum
from pathlib import Path
from typing import Any
from warnings import warn

import xmltodict


class Dataset(object):
    """This is a class for reading and manipulating NcML file.

    Note that NcML documents are used for two distinct purposes:
      - an XML description of NetCDF structure and metadata;
      - create virtual NetCDF datasets, e.g. an aggregation of multiple files.

    This class supports both types of uses.
    """

    def __init__(self, filepath: str = None, location: str = None):
        """

        Parameters
        -----------
        filepath : str
            File path to dataset NcML file. If it does not exist, an empty NcML document will be created and this will
            be the default filename when writing to disk with `to_ncml`.
        location : Str
            NetCDF file location. Set this to create a NcML file modifying an existing NetCDF document.
        """
        self.filepath = Path(filepath) if filepath is not None else None
        if self.filepath and self.filepath.exists():
            # Convert all dictionaries to lists of dicts to simplify the internal logic.
            self.ncroot = self._parse_xml(self.filepath.read_text())

        else:
            self.ncroot = OrderedDict()
            self.ncroot['netcdf'] = OrderedDict(
                {'@xmlns': 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'}
            )
            if location is not None:
                self.ncroot['netcdf']['@location'] = str(location)

    @classmethod
    def from_text(cls, xml: str):
        """Create Dataset from xml string."""
        self = cls()
        self.ncroot = self._parse_xml(xml)
        return self

    @staticmethod
    def _parse_xml(xml: str) -> dict:
        """Return dictionary from xml."""
        return xmltodict.parse(
            xml,
            force_list=['variable', 'attribute', 'group', 'dimension'],
            process_namespaces=True,
            namespaces={
                'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2': None,
                'https://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2': None,
            },
        )

    def __repr__(self):
        return xmltodict.unparse(self.ncroot, pretty=True)

    # Aggregations and scans
    def add_aggregation(
        self, dim_name: str, type_: str, recheck_every: str = None, time_units_change: bool = None
    ):
        """Add aggregation.

        Parameters
        ----------
        dim_name : str
            Dimension name.
        type_ : str
            Aggregation type.
        recheck_every : str
            Time interval for rechecking the aggregation. Only used if `type_` is `AggregationType.scan`.
        time_units_change : bool
            Whether the time units change. Only used if `type_` is `AggregationType.scan`.
        """
        at = AggregationType(type_)
        item = OrderedDict(
            {
                '@dimName': dim_name,
                '@type': at.value,
                '@recheckEvery': recheck_every,
                '@timeUnitsChange': time_units_change,
            }
        )
        item = preparse(item)

        aggregations = self.ncroot['netcdf'].get('aggregation', [])
        for agg in aggregations:
            if agg['@dimName'] == dim_name:
                agg.update(item)
                break
        else:
            aggregations.append(item)
            self.ncroot['netcdf']['aggregation'] = aggregations

    def add_variable_agg(self, dim_name: str, name: str):
        """Add variable aggregation.

        Parameters
        ----------
        dim_name: str
            Dimension name for the aggregation.
        name : str
            Variable name.
        """
        item = OrderedDict({'@name': name})
        aggregations = self.ncroot['netcdf'].get('aggregation')
        for agg in aggregations:
            if agg['@dimName'] == dim_name:
                variables = agg.get('variableAgg', [])
                for var in variables:
                    if var['@name'] == name:
                        var.update(item)
                        break
                else:
                    variables.append(item)
                    agg['variableAgg'] = variables

    def add_scan(
        self,
        dim_name: str,
        location: str,
        reg_exp: str = None,
        suffix: str = None,
        subdirs: bool = True,
        older_than: str = None,
        date_format_mark: str = None,
        enhance: bool = None,
    ):
        """
        Add scan element.

        Parameters
        ----------
        dim_name : str
            Dimension name.
        location : str
            Location of the files to scan.
        reg_exp : str
            Regular expression to match the full pathname of files.
        suffix : str
            File suffix.
        subdirs : bool
            Whether to scan subdirectories.
        older_than : str
            Older than time interval.
        date_format_mark : str
            Date format mark.
        enhance : bool
            Whether to enhance the scan.
        """
        item = OrderedDict(
            {
                '@location': location,
                '@regExp': reg_exp,
                '@suffix': suffix,
                '@subdirs': subdirs,
                '@olderThan': older_than,
                '@dateFormatMark': date_format_mark,
                '@enhance': enhance,
            }
        )

        item = preparse(item)

        # An aggregation must exist for the scan to be added.
        for agg in self.ncroot['netcdf'].get('aggregation'):
            if agg['@dimName'] == dim_name:
                scan = agg.get('scan', [])
                scan.append(item)
                agg['scan'] = scan
                break
        else:
            raise ValueError(f'No aggregation found for dimension {dim_name}.')

    # Variable
    def add_variable_attribute(self, variable, key, value, type_='String'):
        """Add variable attribute.

        Parameters
        ----------
        variable : str
            Variable name
        key : str
            Attribute name
        value : object
            Attribute value. Must be a serializable Python Object
        type_ : str, default: 'String'
             String describing attribute type.

        """
        item = OrderedDict({'@name': key, '@type': type_, '@value': value})
        variables = self.ncroot['netcdf'].get('variable', [])

        for var in variables:
            if var['@name'] == variable:
                var_attributes = var.get('attribute', [])
                for attr in var_attributes:
                    if attr['@name'] == key:
                        attr.update(item)
                        break
                else:
                    var_attributes.append(item)
                    var['attribute'] = var_attributes
                break
        else:
            variables.append(OrderedDict({'@name': variable, 'attribute': item}))
            self.ncroot['netcdf']['variable'] = variables

    def remove_variable_attribute(self, variable, key):
        """Remove variable attribute"""
        item = OrderedDict({'@name': key, '@type': 'attribute'})
        variables = self.ncroot['netcdf'].get('variable', [])

        for var in variables:
            if var['@name'] == variable:
                var['remove'] = item
                break
        else:
            new_var = OrderedDict({'@name': variable, 'remove': item})
            variables.append(new_var)
            self.ncroot['netcdf']['variable'] = variables

    def rename_variable(self, variable, new_name):
        """Rename variable attribute

        Parameters
        ----------
        variable : str
            Original variable name.
        new_name : str
            New variable name.

        """
        item = OrderedDict({'@name': new_name, '@orgName': variable})
        variables = self.ncroot['netcdf'].get('variable', [])

        for var in variables:
            if var['@name'] == variable:
                var['@name'] = new_name
                var['@orgName'] = variable
                break
        else:
            variables.append(item)
            self.ncroot['netcdf']['variable'] = variables

    def remove_variable(self, variable):
        """Remove dataset variable.

        Parameters
        ----------
        key : str
            Name of the variable to remove.
        """
        item = OrderedDict({'@name': variable, '@type': 'variable'})
        removes = self.ncroot['netcdf'].get('remove', [])

        if item not in removes:
            removes.append(item)
            self.ncroot['netcdf']['remove'] = removes

    def rename_variable_attribute(self, variable, old_name, new_name):
        """Rename variable attribute.

        Parameters
        ----------
        variable : str
          Variable name.
        old_name : str
          Original attribute name.
        new_name : str
          New attribute name.
        """
        item = OrderedDict({'@name': new_name, '@orgName': old_name})
        variables = self.ncroot['netcdf'].get('variable', [])

        for var in variables:
            if var['@name'] == variable:
                attrs = var.get('attribute', [])
                for attr in attrs:
                    if attr['@name'] == old_name:
                        attr['@name'] = new_name
                        attr['@orgName'] = old_name
                        break
                else:
                    attrs.append(item)
                    break
        else:
            new_var = OrderedDict({'@name': 'variable', 'attribute': item})
            variables.append(new_var)
            self.ncroot['netcdf']['variable'] = variables

    # Dimensions

    def rename_dimension(self, dimension, new_name):
        """Rename dimension.

        Parameters
        ----------
        dimension: str
          Original dimension name.
        new_name: str
          New dimension name.
        """
        item = OrderedDict({'@name': new_name, '@orgName': dimension})
        dimensions = self.ncroot['netcdf'].get('dimension', [])

        for dim in dimensions:
            if dim['@name'] == dimension:
                dim['@name'] = new_name
                dim['@orgName'] = dimension
                break
        else:
            dimensions.append(item)
            self.ncroot['netcdf']['dimensions'] = dimensions

    # Dataset

    def add_dataset_attribute(self, key, value, type_='String'):
        """Add dataset attribute
         Parameters
        ----------
        key : str
            Attribute name.
        value : object
            Attribute value. Must be a serializable Python Object.
        type_ : str, default: 'String'
            String describing attribute type.

        """
        item = OrderedDict({'@name': key, '@type': type_, '@value': value})
        attributes = self.ncroot['netcdf'].get('attribute', [])

        for attr in attributes:
            if attr['@name'] == key:
                attr.update(item)
                break
        else:
            attributes.append(item)
            self.ncroot['netcdf']['attribute'] = attributes

    def remove_dataset_attribute(self, key):
        """Remove dataset attribute.

        Parameters
        ----------
        key : str
            Name of the attribute to remove.

        """
        removals = self.ncroot['netcdf'].get('remove', [])
        item = OrderedDict({'@name': key, '@type': 'attribute'})

        if removals:
            removals_keys = [
                removal['@name'] for removal in removals if removal['@type'] == 'attribute'
            ]
            if key not in removals_keys:
                removals.append(item)
        else:
            self.ncroot['netcdf']['remove'] = [item]

    def rename_dataset_attribute(self, old_name, new_name):
        """Rename dataset attribute.

        Parameters
        ----------
        old_name: str
          Original attribute name.
        new_name: str
          New attribute name.
        """

        attributes = self.ncroot['netcdf'].get('attribute', None)
        item = OrderedDict({'@name': new_name, 'orgName': old_name})

        if attributes:
            if isinstance(attributes, (dict, OrderedDict)):
                attributes = [attributes]

            for attr in attributes:
                if attr['@name'] == old_name:
                    attr['@name'] = new_name
                    attr['@orgName'] = old_name
                    break
            else:
                self.ncroot['netcdf']['attribute'] = [*attributes, item]

        else:
            self.ncroot['netcdf']['attribute'] = item

    def to_ncml(self, path=None):
        """Write NcML file to disk.

        Parameters
        ----------
        path: str
          Path to write NcML document.
        """
        if not path:
            if self.filepath.exists():
                path = f'{str(self.filepath).strip(".ncml")}_modified.ncml'
            else:
                path = str(self.filepath)

        xml_output = xmltodict.unparse(self.ncroot, pretty=True)
        with open(path, 'w') as fd:
            fd.write(xml_output)

    def to_cf_dict(self):
        """Convert internal representation to a CF-JSON dictionary.

        The CF-JSON specification includes `data` for variables, but if the data is not within the NcML,
        it cannot be included in the JSON representation.

        Returns
        -------
        Dictionary with `dimensions` and `variables` keys. May also optionally include an `attributes` key and a
        `groups` key. Additional keys prefixed with `@` may be included for <netcdf> tag attributes,
        for example `@location`.

        References
        ----------
        http://cf-json.org/specification
        """
        res = OrderedDict()
        nc = self.ncroot['netcdf']

        for key, val in nc.items():
            if key[0] == '@':
                res[key] = val
            if key == 'dimension':
                res.update(_dims_to_json(val))
            if key == 'group':
                res.update(_groups_to_json(val))
            if key == 'attribute':
                res.update(_attributes_to_json(val))
            if key == 'variable':
                res.update(_variables_to_json(val))

        return res


def _dims_to_json(dims: list) -> dict:
    """The dimensions object has dimension id:size as its key:value members."""
    out = OrderedDict()
    for dim in dims:
        if int(dim['@length']) > 1:
            out[dim['@name']] = int(dim['@length'])

    return {'dimensions': out}


def _groups_to_json(groups: list) -> dict:
    out = OrderedDict()
    for group in groups:
        name = group['@name']
        out[name] = OrderedDict()
        if 'attribute' in group:
            out[name].update(_attributes_to_json(group['attribute']))
        if 'group' in group:
            out[name].update(_groups_to_json(group['group']))

    return {'groups': out}


def _attributes_to_json(attrs: list) -> dict:
    """The attributes object contains arbitrary attributes as its key:value members."""
    out = OrderedDict()
    for attr in attrs:
        try:
            out[attr['@name']] = _cast(attr)
        except ValueError as exc:
            warn(f"Could not cast {attr['@name']}:\n{exc}")

    return {'attributes': out}


def _variables_to_json(variables: list) -> dict:
    """The variables definition object has variable id:object as its key:value members.

    Each variable object MUST include shape, attributes and data objects.
    The shape field is an array of dimension IDs which correspond to the array ordering of the variable data.
    """
    out = OrderedDict()

    # Put coordinate variables first
    for var in variables:
        if _is_coordinate(var):
            out[var['@name']] = None

    for var in variables:
        name = var['@name']
        out[name] = OrderedDict()

        if '@shape' in var:
            out[name]['shape'] = var['@shape'].split(' ')

        if '@type' in var:
            out[name]['type'] = var['@type']

        if 'attribute' in var:
            out[name].update(_attributes_to_json(var['attribute']))

        if 'values' in var:
            out[name]['data'] = _cast(var)

    return {'variables': out}


def _cast(obj: dict) -> Any:
    """Cast attribute value to the appropriate type."""
    from xncml.parser import DataType, nctype

    value = obj.get('@value') or obj.get('values')
    typ = DataType(obj.get('@type', 'String'))
    if value is not None:
        if isinstance(value, str):
            if typ in [DataType.STRING, DataType.STRING_1]:
                return value

            sep = ' '
            values = value.split(sep)
            return list(map(nctype(typ), values))
        elif isinstance(value, dict):
            raise NotImplementedError(obj)
        else:
            return value


def _is_coordinate(var):
    """Return True is variable is a coordinate."""

    # Variable is 1D and has same name as dimension
    if var.get('@shape', '').split(' ') == [var['@name']]:
        return True

    lat_units = ['degrees_north', 'degreeN', 'degree_N', 'degree_north', 'degreesN', 'degrees_N']
    lon_units = ['degrees_east', 'degreeE', 'degree_E', 'degree_east', 'degreesE', 'degrees_E']
    names = [
        'latitude',
        'longitude',
        'time',
        'air_pressure',
        'altitude',
        'depth',
        'geopotential_height',
        'height',
        'height_above_geopotential_datum',
        'height_above_mean_sea_level',
        'height_above_reference_ellipsoid',
    ]

    if 'attribute' in var:
        attrs = _attributes_to_json(var['attribute'])

        # Check units
        if attrs.get('units', '') in lon_units + lat_units:
            return True

        # Check long_name and standard_name
        if attrs.get('long_name', attrs.get('standard_name', '')) in names:
            return True

    return False


def preparse(obj: dict) -> dict:
    """
    - Remove None values from dictionary.
    - Convert booleans to strings.
    """
    for k, v in obj.items():
        if isinstance(v, bool):
            obj[k] = str(v).lower()
    return {k: v for k, v in obj.items() if v is not None}


class AggregationType(Enum):
    """Type of aggregation."""

    FORECAST_MODEL_RUN_COLLECTION = 'forecastModelRunCollection'
    FORECAST_MODEL_RUN_SINGLE_COLLECTION = 'forecastModelRunSingleCollection'
    JOIN_EXISTING = 'joinExisting'
    JOIN_NEW = 'joinNew'
    TILED = 'tiled'
    UNION = 'union'



#!/usr/bin/env python
"""Top-level module for xncml."""
from pkg_resources import DistributionNotFound, get_distribution

from .core import Dataset
from .parser import open_ncml

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass
finally:
    del get_distribution, DistributionNotFound



from .ncml_2_2 import (
    Aggregation,
    AggregationType,
    Attribute,
    CacheVariable,
    DataType,
    Dimension,
    EnumTypedef,
    Group,
    LogicalReduce,
    LogicalSection,
    LogicalSlice,
    Netcdf,
    ObjectType,
    PromoteGlobalAttribute,
    Remove,
    Values,
    Variable,
)

__all__ = [
    'AggregationType',
    'DataType',
    'ObjectType',
    'Aggregation',
    'Attribute',
    'CacheVariable',
    'Dimension',
    'EnumTypedef',
    'Group',
    'LogicalReduce',
    'LogicalSection',
    'LogicalSlice',
    'Netcdf',
    'PromoteGlobalAttribute',
    'Remove',
    'Values',
    'Variable',
]



"""
# NcML parser for xarray

The `open_ncml` function parse an XML document compliant with the NcML-2.2 schema and returns an xarray Dataset.

The XML is parsed into a Python objects using `xsdata`. The parser converts XML elements into classes instances
defined in an autogenerated data model (`generated.ncml_2_2`). This datamodel was created using:

  ```bash
  xsdata generate -ds NumPy --compound-fields -mll=119 --postponed-annotations  schemas/ncml-2.2.xsd
  ```

The code below converts the NcML instructions into xarray instructions. Not all NcML instructions are currently
supported.


## TODO

Support for these elements is missing:
- <cacheVariable>
- <logicalReduce>
- <logicalSection>
- <logicalSlice>
- <promoteGlobalAttribute>
- <enumTypedef>
- <scanFmrc>

Support for these attributes is missing:
- dateFormatMark
- olderThan
- tiled aggregations
"""


from __future__ import annotations

import datetime as dt
from functools import partial
from itertools import chain
from pathlib import Path

import numpy as np
import xarray as xr
from xsdata.formats.dataclass.parsers import XmlParser

from .generated import (
    Aggregation,
    AggregationType,
    Attribute,
    DataType,
    Dimension,
    EnumTypedef,
    Group,
    Netcdf,
    ObjectType,
    Remove,
    Values,
    Variable,
)

__author__ = 'David Huard'
__date__ = 'July 2022'
__contact__ = 'huard.david@ouranos.ca'


def parse(path: Path) -> Netcdf:
    """
    Parse NcML file using NetCDF datamodel based on NcML-2.2 Schema.

    Parameters
    ----------
    path : Path
      Path to NcML file.

    Returns
    -------
    Netcdf instance.
      Object description of NcML content.
    """
    parser = XmlParser()
    return parser.from_path(path, Netcdf)


def open_ncml(ncml: str | Path) -> xr.Dataset:
    """
    Convert NcML document to a dataset.

    Parameters
    ----------
    ncml : str | Path
      Path to NcML file.

    Returns
    -------
    xr.Dataset
      Dataset holding variables and attributes defined in NcML document.
    """
    # Parse NcML document
    ncml = Path(ncml)
    obj = parse(ncml)

    return read_netcdf(xr.Dataset(), xr.Dataset(), obj, ncml)


def read_netcdf(target: xr.Dataset, ref: xr.Dataset, obj: Netcdf, ncml: Path) -> xr.Dataset:
    """
    Return content of <netcdf> element.

    Parameters
    ----------
    target : xr.Dataset
      Target dataset to be updated with <netcdf>'s content.
    ref : xr.Dataset
      Reference dataset used to copy content into `target`.
    obj : Netcdf
       <netcdf> object description.
    ncml : Path
      Path to NcML document, sometimes required to follow relative links.

    Returns
    -------
    xr.Dataset
      Dataset holding variables and attributes defined in <netcdf> element.
    """
    # Open location if any
    ref = read_ds(obj, ncml) or ref

    # <explicit/> element means that only content specifically mentioned in NcML document is included in dataset.
    if obj.explicit is not None:
        pass
    else:
        # By default, all metadata from the reference dataset is read: <readMetadata/>
        target = ref

    for item in filter_by_class(obj.choice, Aggregation):
        target = read_aggregation(target, item, ncml)

    # Handle <variable>, <attribute> and <remove> elements
    target = read_group(target, ref, obj)

    return target


def read_aggregation(target: xr.Dataset, obj: Aggregation, ncml: Path) -> xr.Dataset:
    """
    Return merged or concatenated content of <aggregation> element.

    Parameters
    ----------
    target : xr.Dataset
      Target dataset to be updated with <netcdf>'s content.
    obj : Aggregation
       <aggregation> object description.
    ncml : Path
      Path to NcML document, sometimes required to follow relative links.

    Returns
    -------
    xr.Dataset
      Dataset holding variables and attributes defined in <aggregation> element.
    """
    from xarray.coding.times import CFDatetimeCoder

    # Names of variables to be aggregated. All variables if undefined.
    names = [v.name for v in filter_by_class(obj.choice, Aggregation.VariableAgg)]

    for attr in obj.promote_global_attribute:
        raise NotImplementedError

    # Create list of datasets to aggregate.
    datasets = []
    closers = []

    for item in obj.netcdf:
        # Open dataset defined in <netcdf>'s `location` attribute
        tar = read_netcdf(xr.Dataset(), ref=xr.Dataset(), obj=item, ncml=ncml)
        closers.append(getattr(tar, '_close'))

        # Select variables
        if names:
            tar = tar[names]

        # Handle coordinate values
        if item.coord_value is not None:
            dtypes = [i[obj.dim_name].dtype.type for i in [tar, target] if obj.dim_name in i]
            coords = read_coord_value(item, obj, dtypes=dtypes)
            tar = tar.assign_coords({obj.dim_name: coords})
        datasets.append(tar)

    # Handle <scan> element
    for item in obj.scan:
        dss = read_scan(item, ncml)
        datasets.extend([ds.chunk() for ds in dss])
        closers.extend([getattr(ds, '_close') for ds in dss])

    # Need to decode time variable
    if obj.time_units_change:
        for i, ds in enumerate(datasets):
            t = xr.as_variable(ds[obj.dim_name], obj.dim_name)  # Maybe not the same name...
            encoded = CFDatetimeCoder(use_cftime=True).decode(t, name=t.name)
            datasets[i] = ds.assign_coords({obj.dim_name: encoded})

    # Translate different types of aggregation into xarray instructions.
    if obj.type == AggregationType.JOIN_EXISTING:
        agg = xr.concat(datasets, obj.dim_name)
    elif obj.type == AggregationType.JOIN_NEW:
        agg = xr.concat(datasets, obj.dim_name)
    elif obj.type == AggregationType.UNION:
        agg = xr.merge(datasets)
    else:
        raise NotImplementedError

    agg = read_group(agg, None, obj)
    out = target.merge(agg, combine_attrs='no_conflicts')
    out.set_close(partial(_multi_file_closer, closers))
    return out


def read_ds(obj: Netcdf, ncml: Path) -> xr.Dataset:
    """
    Return dataset defined in <netcdf> element.

    Parameters
    ----------
    obj : Netcdf
      <netcdf> object description.
    ncml : Path
      Path to NcML document, sometimes required to follow relative links.

    Returns
    -------
    xr.Dataset
      Dataset defined at <netcdf>' `location` attribute.
    """
    if obj.location:
        try:
            # Python >= 3.9
            location = obj.location.removeprefix('file:')
        except AttributeError:
            location = obj.location.strip('file:')

        if not Path(location).is_absolute():
            location = ncml.parent / location
        return xr.open_dataset(location, decode_times=False)


def read_group(target: xr.Dataset, ref: xr.Dataset, obj: Group | Netcdf) -> xr.Dataset:
    """
    Parse <group> items, typically <dimension>, <variable>, <attribute> and <remove> elements.

    Parameters
    ----------
    target : xr.Dataset
      Target dataset to be updated.
    ref : xr.Dataset
      Reference dataset used to copy content into `target`.
    obj : Group | Netcdf
       <netcdf> object description.

    Returns
    -------
    xr.Dataset
      Dataset holding variables and attributes defined in <netcdf> element.
    """
    dims = {}
    for item in obj.choice:
        if isinstance(item, Dimension):
            dims[item.name] = read_dimension(item)
        elif isinstance(item, Variable):
            target = read_variable(target, ref, item, dims)
        elif isinstance(item, Attribute):
            read_attribute(target, item, ref)
        elif isinstance(item, Remove):
            target = read_remove(target, item)
        elif isinstance(item, EnumTypedef):
            raise NotImplementedError
        elif isinstance(item, Group):
            target = read_group(target, ref, item)
        elif isinstance(item, Aggregation):
            pass  # <aggregation> elements are parsed in `read_netcdf`
        else:
            raise AttributeError

    return target


def read_scan(obj: Aggregation.Scan, ncml: Path) -> [xr.Dataset]:
    """
    Return list of datasets defined in <scan> element.

    Parameters
    ----------
    obj : Aggregation.Scan instance
      <scan> object description.
    ncml : Path
      Path to NcML document, sometimes required to follow relative links.

    Returns
    -------
    list
      List of datasets found by scan.
    """
    import glob
    import re

    if obj.date_format_mark:
        raise NotImplementedError

    path = Path(obj.location)
    if not path.is_absolute():
        path = ncml.parent / path

    files = list(path.rglob('*') if obj.subdirs else path.glob('*'))

    if not files:
        raise ValueError(f'No files found in {path}')

    fns = map(str, files)
    if obj.reg_exp:
        pat = re.compile(obj.reg_exp)
        files = list(filter(pat.match, fns))
    elif obj.suffix:
        pat = '*' + obj.suffix
        files = glob.fnmatch.filter(fns, pat)

    if not files:
        raise ValueError('regular expression or suffix matches no file.')

    files.sort()

    return [xr.open_dataset(f, decode_times=False) for f in files]


def read_coord_value(nc: Netcdf, agg: Aggregation, dtypes: list = ()):
    """
    Read `coordValue` attribute of <netcdf> element.

    Parameters
    ----------
    nc : Netcdf instance
      <netcdf> object description.
    agg : Aggregation instance
      <aggregation> object description
    dtypes : tuple
      List of preferred type for coordinate value.

    Returns
    -------
    str, np.array, scalar
      Coordinate value cast to proper type.

    Notes
    -----
    The casting logic is most likely not up to spec.
    """
    val = nc.coord_value

    # A JOIN_NEW aggregation has exactly one coordinate value
    if agg.type == AggregationType.JOIN_NEW:
        coord = val
    elif agg.type == AggregationType.JOIN_EXISTING:
        coord = val.replace(',', ' ').split()
    else:
        raise NotImplementedError

    # Cast to dtype, not clear what the spec is exactly for this.
    if dtypes:
        typ = dtypes[0]
    else:
        try:
            dt.datetime.strptime(coord, '%Y-%m-%d %H:%M:%SZ')
            typ = str
        except ValueError:
            typ = np.float64

    return typ(coord)


def read_variable(target: xr.Dataset, ref: xr.Dataset, obj: Variable, dimensions: dict):
    """
    Parse <variable> element.

    Parameters
    ----------
    target : xr.Dataset
      Target dataset to be updated.
    ref : xr.Dataset
      Reference dataset used to copy content into `target`.
    obj : Variable
       <variable> object description.
    dimensions : dict
      Dimension attributes keyed by name.

    Returns
    -------
    xr.Dataset
      Dataset holding variable defined in <variable> element.
    """
    # Handle logic for variable name change
    if obj.org_name:
        if obj.org_name in target:
            target = target.rename({obj.org_name: obj.name})
            ref_var = None
        elif obj.org_name in ref:
            ref_var = xr.as_variable(ref[obj.org_name])
        else:
            raise ValueError
    else:
        ref_var = None

    # Read existing data or create empty DataArray
    if (obj.name in target) or (obj.name in target.dims):
        out = xr.as_variable(target[obj.name])
        if obj.type:
            out = out.astype(nctype(obj.type))
        ref_var = None
    elif (obj.name in ref) or (obj.name in ref.dims):
        out = xr.as_variable(ref[obj.name])
        if obj.type:
            out = out.astype(nctype(obj.type))
        ref_var = ref[obj.name]
    elif obj.shape:
        dims = obj.shape.split(' ')
        shape = [dimensions[dim].length for dim in dims]
        out = xr.Variable(data=np.empty(shape, dtype=nctype(obj.type)), dims=dims)
    else:
        raise ValueError

    # Set variable attributes
    for item in obj.attribute:
        read_attribute(out, item, ref=ref_var)

    # Remove attributes or dimensions
    for item in obj.remove:
        read_remove(out, item)

    # Read values
    if obj.values:
        out = read_values(out, obj.values)

    if obj.logical_section:
        raise NotImplementedError

    if obj.logical_slice:
        raise NotImplementedError

    if obj.logical_reduce:
        raise NotImplementedError

    if obj.typedef:
        raise NotImplementedError

    target[obj.name] = out
    return target


def read_values(v: xr.Variable, obj: Values) -> xr.Variable:
    """Read values for <variable> element.

    Parameters
    ----------
    v : xr.DataArray
      Array whose values are to be updated.
    obj : Values instance
      <values> object description

    Returns
    -------
    xr.Variable
      Array filled with values from <values> element.
    """
    if obj.from_attribute is not None:
        raise NotImplementedError

    n = int(obj.npts or v.size)
    if obj.start is not None and obj.increment is not None:
        data = obj.start + np.arange(n) * obj.increment
    else:
        sep = obj.separator or ' '
        if isinstance(obj.content, list) and isinstance(obj.content[0], str):
            data = obj.content[0].split(sep)
        else:
            raise NotImplementedError

    data = v.dtype.type(data)
    return xr.Variable(v.dims, data, v.attrs)


def read_remove(target: xr.Dataset | xr.Variable, obj: Remove) -> xr.Dataset:
    """Remove item from dataset.

    Parameters
    ----------
    target : xr.Dataset | xr.Variable
      Target dataset or variable to be updated.
    obj : Remove instance
      <remove> object description.

    Returns
    -------
    xr.Dataset or xr.Variable
      Dataset with attribute, variable or dimension removed, or variable with attribute removed.
    """

    if obj.type == ObjectType.ATTRIBUTE:
        target.attrs.pop(obj.name)
    elif obj.type == ObjectType.VARIABLE:
        target = target.drop_vars(obj.name)
    elif obj.type == ObjectType.DIMENSION:
        target = target.drop_dims(obj.name)

    return target


def read_attribute(target: xr.Dataset | xr.Variable, obj: Attribute, ref: xr.Dataset = None):
    """Update target dataset in place with new or modified attribute.

    Parameters
    ----------
    target : xr.Dataset | xr.Variable
      Target dataset to be updated.
    obj : Attribute instance
      <attribute> object description.
    ref : xr.Dataset
      Reference dataset.
    """
    if obj.value is not None:
        target.attrs[obj.name] = cast(obj)
    elif obj.org_name is not None:
        try:
            target.attrs[obj.name] = ref.attrs[obj.org_name]
        except (AttributeError, KeyError):
            target.attrs[obj.name] = target.attrs.get(obj.org_name)
        target.attrs.pop(obj.org_name)
    elif ref is not None:
        target.attrs[obj.name] = ref.attrs.get(obj.name)
    else:
        raise NotImplementedError


def read_dimension(obj: Dimension) -> Dimension:
    """Return dimension object with its length cast to an integer."""
    if obj.length is not None:
        obj.length = int(obj.length)

    return obj


def nctype(typ: DataType) -> type:
    """Return Python type corresponding to the NcML DataType of object."""

    if typ in [DataType.STRING, DataType.STRING_1]:
        return str
    elif typ == DataType.BYTE:
        return np.int8
    elif typ == DataType.SHORT:
        return np.int16
    elif typ == DataType.INT:
        return np.int32
    elif typ == DataType.LONG:
        return int
    elif typ == DataType.FLOAT:
        return np.float32
    elif typ == DataType.DOUBLE:
        return np.float64

    raise NotImplementedError


def cast(obj: Attribute):
    """Cast attribute value to the appropriate type."""
    value = obj.value or obj.content
    if value:
        if obj.type in [DataType.STRING, DataType.STRING_1]:
            return value

        sep = obj.separator or ' '
        values = value.split(sep)
        return tuple(map(nctype(obj.type), values))


def filter_by_class(iterable, klass):
    """Return generator filtering on class."""
    for item in iterable:
        if isinstance(item, klass):
            yield item


def _multi_file_closer(closers):
    """Close multiple files."""
    # Note that if a closer is None, it probably means an alteration was made to the original dataset. Make sure
    # that the `_close` attribute is obtained directly from the object returned by `open_dataset`.
    for closer in closers:
        if closer is not None:
            closer()



import os
import tempfile
from collections import OrderedDict
from pathlib import Path

import pytest

import xncml

here = os.path.abspath(os.path.dirname(__file__))
input_file = Path(here) / 'data' / 'exercise1.ncml'


def test_ncml_dataset_constructor():
    # Test with existing NcML
    nc = xncml.Dataset(input_file)
    expected = OrderedDict(
        [
            ('@name', 'T'),
            ('@shape', 'time lat lon'),
            ('@type', 'double'),
            (
                'attribute',
                [
                    OrderedDict(
                        [
                            ('@name', 'long_name'),
                            ('@type', 'String'),
                            ('@value', 'surface temperature'),
                        ]
                    ),
                    OrderedDict([('@name', 'units'), ('@type', 'String'), ('@value', 'C')]),
                ],
            ),
        ]
    )
    res = nc.ncroot['netcdf']['variable'][1]
    assert res == expected

    # Test with non-existing NcML
    nc = xncml.Dataset('example.ncml')
    assert '@xmlns' in nc.ncroot['netcdf']

    # Test with non-exising NcML and location
    nc = xncml.Dataset('example.ncml', location=Path(here) / 'data' / 'nc' / 'example1.nc')
    assert 'example1.nc' in nc.ncroot['netcdf']['@location']

    # Test with namespace
    nc = xncml.Dataset(Path(here) / 'data' / 'testReadHttps.xml')
    assert nc.ncroot['netcdf']['attribute'][0]['@value'] == 'Example Data'


def test_add_variable_attribute():
    nc = xncml.Dataset(input_file)
    nc.add_variable_attribute(variable='T', key='container', value='ndarray')
    expected = OrderedDict(
        [
            ('@name', 'T'),
            ('@shape', 'time lat lon'),
            ('@type', 'double'),
            (
                'attribute',
                [
                    OrderedDict(
                        [
                            ('@name', 'long_name'),
                            ('@type', 'String'),
                            ('@value', 'surface temperature'),
                        ]
                    ),
                    OrderedDict([('@name', 'units'), ('@type', 'String'), ('@value', 'C')]),
                    OrderedDict(
                        [('@name', 'container'), ('@type', 'String'), ('@value', 'ndarray')]
                    ),
                ],
            ),
        ]
    )

    res = nc.ncroot['netcdf']['variable'][1]
    assert res == expected

    nc.add_variable_attribute(variable='Tasmax', key='units', value='kelvin')
    res = nc.ncroot['netcdf']['variable'][5]
    expected = OrderedDict(
        [
            ('@name', 'Tasmax'),
            (
                'attribute',
                OrderedDict([('@name', 'units'), ('@type', 'String'), ('@value', 'kelvin')]),
            ),
        ]
    )
    assert res == expected


@pytest.mark.parametrize(
    'variable,key,expected, var_index',
    [
        (
            'T',
            'units',
            OrderedDict(
                [
                    ('@name', 'T'),
                    ('@shape', 'time lat lon'),
                    ('@type', 'double'),
                    (
                        'attribute',
                        [
                            OrderedDict(
                                [
                                    ('@name', 'long_name'),
                                    ('@type', 'String'),
                                    ('@value', 'surface temperature'),
                                ]
                            ),
                            OrderedDict([('@name', 'units'), ('@type', 'String'), ('@value', 'C')]),
                        ],
                    ),
                    ('remove', OrderedDict([('@name', 'units'), ('@type', 'attribute')])),
                ]
            ),
            1,
        ),
        (
            'Tidi',
            'unwantedvaribleAttribute',
            OrderedDict(
                [
                    ('@name', 'Tidi'),
                    (
                        'remove',
                        OrderedDict(
                            [('@name', 'unwantedvaribleAttribute'), ('@type', 'attribute')]
                        ),
                    ),
                ]
            ),
            5,
        ),
    ],
)
def test_remove_variable_attribute(variable, key, expected, var_index):
    nc = xncml.Dataset(input_file)
    nc.remove_variable_attribute(variable=variable, key=key)
    res = nc.ncroot['netcdf']['variable'][var_index]
    assert res == expected


def test_rename_variable():
    # Rename existing variable
    nc = xncml.Dataset(input_file)
    nc.rename_variable('lat', 'latitude')
    res = nc.ncroot['netcdf']['variable'][2]
    expected = OrderedDict(
        [
            ('@name', 'latitude'),
            ('@shape', 'lat'),
            ('@type', 'float'),
            (
                'attribute',
                [
                    OrderedDict(
                        [('@name', 'units'), ('@type', 'String'), ('@value', 'degrees_north')]
                    ),
                ],
            ),
            ('values', '41.0 40.0 39.0'),
            ('@orgName', 'lat'),
        ]
    )

    assert expected == res

    # Rename non-existing variable
    nc.rename_variable('Temp', 'Temperature')
    res = nc.ncroot['netcdf']['variable'][-1]
    assert res == OrderedDict([('@name', 'Temperature'), ('@orgName', 'Temp')])


def test_rename_variable_attribute():
    # Rename existing attribute
    nc = xncml.Dataset(input_file)
    expected = [
        OrderedDict(
            [
                ('@name', 'Units'),
                ('@type', 'String'),
                ('@value', 'degrees_north'),
                ('@orgName', 'units'),
            ]
        )
    ]

    nc.rename_variable_attribute('lat', 'units', 'Units')
    res = nc.ncroot['netcdf']['variable'][2]['attribute']
    assert res == expected

    # Rename non-existing attribute (could be in netCDF file but not in NcML)
    nc.rename_variable_attribute('lat', 'foo', 'bar')
    res = nc.ncroot['netcdf']['variable'][2]['attribute']
    assert {'@name': 'bar', '@orgName': 'foo'} in res


def test_rename_dimension():
    nc = xncml.Dataset(input_file)
    nc.rename_dimension('time', 'Time')
    res = nc.ncroot['netcdf']['dimension']
    expected = [
        OrderedDict(
            [('@name', 'Time'), ('@length', '2'), ('@isUnlimited', 'true'), ('@orgName', 'time')]
        ),
        OrderedDict([('@name', 'lat'), ('@length', '3')]),
        OrderedDict([('@name', 'lon'), ('@length', '4')]),
    ]

    assert res == expected

    # With non-existing dimension
    nc.rename_dimension('time_bound', 'time_bounds')
    assert '@orgName' in res[-1]


def test_add_dataset_attribute():
    nc = xncml.Dataset(input_file)
    nc.add_dataset_attribute(key='editedby', value='foo')
    nc.add_dataset_attribute(key='editedby', value='bar')
    expected = [
        OrderedDict([('@name', 'title'), ('@type', 'String'), ('@value', 'Example Data')]),
        OrderedDict([('@name', 'editedby'), ('@type', 'String'), ('@value', 'bar')]),
    ]
    res = nc.ncroot['netcdf']['attribute']
    assert res == expected


def test_remove_dataset_attribute():
    nc = xncml.Dataset(input_file)
    nc.add_dataset_attribute('bar', 'foo')
    nc.remove_dataset_attribute('title')
    nc.remove_dataset_attribute('title')
    nc.remove_dataset_attribute('bar')
    expected_removals = nc.ncroot['netcdf']['remove']
    expected_removals = [
        removal for removal in expected_removals if removal['@type'] == 'attribute'
    ]
    assert len(expected_removals) == 2


def test_rename_dataset_attribute():
    nc = xncml.Dataset(input_file)
    # Rename existing attribute
    nc.rename_dataset_attribute(old_name='title', new_name='Title')
    assert nc.ncroot['netcdf']['attribute'][0]['@name'] == 'Title'

    # Rename attribute not in the NcML (but possibly in the netcdf `location`)
    nc.rename_dataset_attribute(old_name='foo', new_name='bar')
    assert nc.ncroot['netcdf']['attribute'][1]['@name'] == 'bar'


def test_remove_variable():
    nc = xncml.Dataset(input_file)
    nc.remove_variable('lon')
    expected = [OrderedDict([('@name', 'lon'), ('@type', 'variable')])]
    res = nc.ncroot['netcdf']['remove']
    assert expected == res


def test_add_aggregation():
    nc = xncml.Dataset(input_file)
    nc.add_aggregation('new_dim', 'joinNew')
    nc.add_variable_agg('new_dim', 'newVar')

    expected = [
        OrderedDict(
            [
                ('@dimName', 'new_dim'),
                ('@type', 'joinNew'),
                ('variableAgg', [OrderedDict([('@name', 'newVar')])]),
            ]
        )
    ]
    res = nc.ncroot['netcdf']['aggregation']

    assert expected == res


def test_add_scan():
    nc = xncml.Dataset(input_file)
    nc.add_aggregation('new_dim', 'joinExisting')
    nc.add_scan('new_dim', location='foo', suffix='.nc')

    expected = [
        OrderedDict(
            [
                ('@dimName', 'new_dim'),
                ('@type', 'joinExisting'),
                (
                    'scan',
                    [OrderedDict([('@location', 'foo'), ('@subdirs', 'true'), ('@suffix', '.nc')])],
                ),
            ]
        )
    ]

    res = nc.ncroot['netcdf']['aggregation']
    assert expected == res


def test_to_ncml():
    nc = xncml.Dataset(input_file)
    with tempfile.NamedTemporaryFile(suffix='.ncml') as t:
        nc.to_ncml(path=t.name)
        assert os.path.exists(t.name)

    nc.to_ncml()
    default = f'{str(input_file).strip(".ncml")}_modified.ncml'
    assert os.path.exists(default)
    try:
        os.remove(default)
    except Exception:
        pass


def test_to_dict():
    nc = xncml.Dataset(input_file)
    out = nc.to_cf_dict()
    assert out['attributes']['title'] == 'Example Data'
    assert out['variables']['rh']['attributes']['long_name'] == 'relative humidity'
    assert out['variables']['rh']['type'] == 'int'
    assert out['variables']['rh']['shape'] == ['time', 'lat', 'lon']
    assert out['dimensions']['time'] == 2
    assert 'groups' not in out

    # Check coordinates are first
    assert list(out['variables'].keys())[:3] == ['lat', 'lon', 'time']

    nc = xncml.Dataset(Path(here) / 'data' / 'aggNewCoord.ncml')
    out = nc.to_cf_dict()
    assert out['variables']['time']['data'] == [0, 1, 2]

    nc = xncml.Dataset(Path(here) / 'data' / 'subsetCoordEdges.ncml')
    with pytest.raises(NotImplementedError):
        out = nc.to_cf_dict()


def test_from_xml():
    nc1 = xncml.Dataset(input_file)

    xml = input_file.read_text()
    nc2 = xncml.Dataset.from_text(xml)

    assert str(nc2) == str(nc1)



import datetime as dt
from pathlib import Path

import numpy as np
import psutil
import pytest

import xncml

# Notes

# This is not testing absolute paths.
# Would need to modify the XML files _live_ to reflect the actual path.


data = Path(__file__).parent / 'data'


class CheckClose(object):
    """Check that files are closed after the test. Note that `close` has to be explicitly called within the
    context manager for this to work."""

    def __init__(self):
        self.proc = psutil.Process()
        self.before = None

    def __enter__(self):
        self.before = len(self.proc.open_files())

    def __exit__(self, *args):
        """Raise error if files are left open at the end of the test."""
        after = len(self.proc.open_files())
        if after != self.before:
            raise AssertionError(f'Files left open after test: {after - self.before}')


def test_aggexisting():
    with CheckClose():
        ds = xncml.open_ncml(data / 'aggExisting.xml')
        check_dimension(ds)
        check_coord_var(ds)
        check_agg_coord_var(ds)
        check_read_data(ds)
        assert ds['time'].attrs['ncmlAdded'] == 'timeAtt'
        ds.close()


def test_aggexisting_w_coords():
    with CheckClose():
        ds = xncml.open_ncml(data / 'aggExistingWcoords.xml')
        check_dimension(ds)
        check_coord_var(ds)
        check_agg_coord_var(ds)
        check_read_data(ds)
        assert ds['time'].attrs['ncmlAdded'] == 'timeAtt'
        ds.close()


def test_aggexisting_coords_var():
    ds = xncml.open_ncml(data / 'aggExisting1.xml')
    check_dimension(ds)
    check_coord_var(ds)
    check_agg_coord_var(ds)
    check_read_data(ds)
    assert all(ds['time'].data == list(range(7, 125, 2)))


def test_agg_new():
    ds = xncml.open_ncml(data / 'aggNew.ncml')
    assert len(ds.time) == 3
    assert all(ds.time.data == [0, 10, 99])
    assert 'T' in ds.data_vars
    assert len(ds.lat) == 3


def test_agg_new_coord():
    ds = xncml.open_ncml(data / 'aggNewCoord.ncml')
    assert ds.time.dtype == np.int32
    assert len(ds.time) == 3
    assert all(ds.time.data == [0, 1, 2])
    assert ds.time.attrs['units'] == 'months since 2000-6-16 6:00'
    assert 'T' in ds.data_vars
    assert len(ds.lat) == 3


def test_agg_existing2():
    ds = xncml.open_ncml(data / 'aggExisting2.xml')
    assert ds['time'].attrs['units'] == 'hours since 2006-06-16 00:00'
    assert ds['time'].dtype == float
    assert all(ds['time'].data == [12.0, 13.0, 14.0])


def test_agg_existing4():
    ds = xncml.open_ncml(data / 'aggExisting4.ncml')
    assert all(ds['time'].data == [1.1496816e9, 1.1496852e9, 1.1496888e9])


def test_agg_existing5():
    ds = xncml.open_ncml(data / 'aggExisting5.ncml')
    assert ds['time'].dtype == np.int32
    assert all(ds['time'].data == list(range(59)))


def test_agg_existing_add_coords():
    # TODO: Complete test
    ds = xncml.open_ncml(data / 'aggExistingAddCoord.ncml')
    assert 'time' in ds.variables


def test_modify_atts():
    ds = xncml.open_ncml(data / 'modifyAtts.xml')
    assert ds.attrs['Conventions'] == 'Metapps'
    assert 'title' not in ds.attrs
    assert 'UNITS' in ds['rh'].attrs
    assert 'units' not in ds['rh'].attrs
    assert ds['rh'].attrs['longer_name'] == 'Abe said what?'
    assert 'long_name' not in ds['rh'].attrs


def test_modify_vars():
    ds = xncml.open_ncml(data / 'modifyVars.xml')
    assert ds.attrs['Conventions'] == 'added'
    assert ds.attrs['title'] == 'replaced'

    assert 'deltaLat' in ds.data_vars
    assert all(ds['deltaLat'].data == [0.1, 0.1, 0.01])
    assert ds['deltaLat'].dtype == float

    assert 'Temperature' in ds.data_vars
    assert 'T' not in ds.data_vars

    assert 'ReletiveHumidity' in ds.data_vars
    assert 'rh' not in ds.data_vars
    rh = ds['ReletiveHumidity']
    assert rh.attrs['long_name2'] == 'relatively humid'
    assert rh.attrs['units'] == 'percent (%)'
    assert 'long_name' not in rh.attrs


def test_agg_syn_grid():
    ds = xncml.open_ncml(data / 'aggSynGrid.xml')
    assert len(ds.lat) == 3
    assert len(ds.lon) == 4
    assert len(ds.time) == 3
    assert all(ds.time == ['2005-11-22 22:19:53Z', '2005-11-22 23:19:53Z', '2005-11-23 00:19:59Z'])


def test_agg_syn_no_coord():
    ds = xncml.open_ncml(data / 'aggSynNoCoord.xml')
    assert len(ds.lat) == 3
    assert len(ds.lon) == 4
    assert len(ds.time) == 3


def test_agg_syn_no_coords_dir():
    ds = xncml.open_ncml(data / 'aggSynNoCoordsDir.xml')
    assert len(ds.lat) == 3
    assert len(ds.lon) == 4
    assert len(ds.time) == 3


def test_agg_synthetic():
    ds = xncml.open_ncml(data / 'aggSynthetic.xml')
    assert len(ds.time) == 3
    assert all(ds.time == [0, 10, 99])


def test_agg_synthetic_2():
    ds = xncml.open_ncml(data / 'aggSynthetic2.xml')
    assert len(ds.time) == 3
    assert all(ds.time == [0, 1, 2])


def test_agg_synthetic_3():
    ds = xncml.open_ncml(data / 'aggSynthetic3.xml')
    assert len(ds.time) == 3
    assert all(ds.time == [0, 10, 99])


def test_agg_syn_scan():
    with CheckClose():
        ds = xncml.open_ncml(data / 'aggSynScan.xml')
        assert len(ds.time) == 3
        assert all(ds.time == [0, 10, 20])
        ds.close()


def test_agg_syn_rename():
    ds = xncml.open_ncml(data / 'aggSynRename.xml')
    assert len(ds.time) == 3
    assert 'T' not in ds
    assert 'Temperature' in ds


def test_rename_var():
    ds = xncml.open_ncml(data / 'renameVar.xml')
    assert ds.attrs['title'] == 'Example Data'

    assert 'ReletiveHumidity' in ds
    assert all(ds.lat.data == [41.0, 40.0, 39.0])
    assert all(ds.lon.data == [-109.0, -107.0, -105.0, -103.0])
    assert ds.lon.dtype == np.float32
    assert all(ds.time.data == [6, 18, 24, 36])
    assert ds.time.dtype == np.int32

    assert all(np.equal(ds.attrs['testFloat'], [1.0, 2.0, 3.0, 4.0]))

    assert ds.attrs['testByte'][0].dtype == np.int8
    assert ds.attrs['testShort'][0].dtype == np.int16
    assert ds.attrs['testInt'][0].dtype == np.int32
    assert ds.attrs['testFloat'][0].dtype == np.float32
    assert ds.attrs['testDouble'][0].dtype == np.float64


def test_agg_union_simple():
    ds = xncml.open_ncml(data / 'aggUnionSimple.xml')
    assert ds.attrs['title'] == 'Union cldc and lflx'
    assert len(ds.lat) == 21
    assert ds.lat.attrs['units'] == 'degrees_north'
    assert all(ds.lat.data[:3] == [10, 9, 8])
    assert len(ds.time) == 456
    assert 'lflx' in ds
    assert 'cldc' in ds
    assert ds.lflx.shape == (456, 21, 360)


def test_agg_union():
    ds = xncml.open_ncml(data / 'aggUnion.xml')
    assert ds.attrs['title'] == 'Example Data'
    assert ds.lat.size == 3
    assert ds.time.size == 2
    assert ds.ReletiveHumidity.shape == (2, 3, 4)
    assert ds.ReletiveHumidity.attrs['units'] == 'percent'
    assert ds.Temperature.shape == (2, 3, 4)
    assert ds.Temperature.attrs['units'] == 'degC'


def test_agg_union_rename():
    ds = xncml.open_ncml(data / 'aggUnionRename.xml')
    assert 'LavaFlow' in ds.variables


def test_agg_union_scan():
    ds = xncml.open_ncml(data / 'aggUnionScan.xml')
    assert 'lflx' in ds
    assert 'cldc' in ds


def test_read():
    ds = xncml.open_ncml(data / 'testRead.xml')
    assert ds.attrs['title'] == 'Example Data'
    assert ds.attrs['testFloat'] == (1.0, 2.0, 3.0, 4.0)


def test_read_override():
    ds = xncml.open_ncml(data / 'testReadOverride.xml')
    assert 'rh' not in ds.variables


@pytest.mark.skip(reason='unclear if this is meant to fail')
def test_read_https():
    ds = xncml.open_ncml(data / 'testReadHttps.xml')
    assert ds.attrs['title'] == 'Example Data'


def test_agg_existing_inequivalent_cals():
    ds = xncml.open_ncml(data / 'agg_with_calendar/aggExistingInequivalentCals.xml')
    assert ds.time.size == 725
    assert ds.time[-1] == dt.datetime(2018, 12, 31)


@pytest.mark.skip(reason='dateFormatMark not implemented')
def test_aggexistingone():
    ds = xncml.open_ncml(data / 'aggExistingOne.ncml')
    assert len(ds.time) == 3


@pytest.mark.skip(reason='dateFormatMark not implemented')
@pytest.mark.skip(reason='<promoteGlobalAttribute> not implemented')
def test_agg_existing_promote():
    ds = xncml.open_ncml(data / 'aggExistingPromote.ncml')
    assert 'times' in ds.variables


@pytest.mark.skip(reason='<promoteGlobalAttribute> not implemented')
def test_agg_existing_promote2():
    _ = xncml.open_ncml(data / 'aggExistingPromote2.ncml')


def test_agg_join_new_scalar_coord():
    _ = xncml.open_ncml(data / 'aggJoinNewScalarCoord.xml')
    # TODO: Complete test


def test_exercise_1():
    _ = xncml.open_ncml(data / 'exercise1.ncml')
    # TODO: Complete test


def test_read_meta_data():
    ds = xncml.open_ncml(data / 'readMetadata.xml')
    assert ds.attrs['title'] == 'Example Data'
    assert ds.variables['T'].attrs['units'] == 'degC'


# --- #
def check_dimension(ds):
    assert len(ds['lat']) == 3
    assert len(ds['lon']) == 4
    assert len(ds['time']) == 59


def check_coord_var(ds):
    lat = ds['lat']
    assert len(lat) == 3
    assert lat.dtype == np.float32
    assert lat.dims == ('lat',)
    assert lat.attrs['units'] == 'degrees_north'
    assert all(lat.data == [41.0, 40.0, 39.0])


def check_agg_coord_var(ds):
    time = ds['time']
    assert time.dims == ('time',)
    assert len(time) == 59
    assert time.dtype == np.int32


def check_read_data(ds):
    t = ds['T']
    assert t.dims == ('time', 'lat', 'lon')
    assert t.size == 708
    assert t.shape == (59, 3, 4)
    assert t.dtype == float
    assert 'T' in ds.data_vars



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Type

__NAMESPACE__ = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'


class AggregationType(Enum):
    FORECAST_MODEL_RUN_COLLECTION = 'forecastModelRunCollection'
    FORECAST_MODEL_RUN_SINGLE_COLLECTION = 'forecastModelRunSingleCollection'
    JOIN_EXISTING = 'joinExisting'
    JOIN_NEW = 'joinNew'
    TILED = 'tiled'
    UNION = 'union'


class DataType(Enum):
    BYTE = 'byte'
    CHAR = 'char'
    SHORT = 'short'
    INT = 'int'
    LONG = 'long'
    FLOAT = 'float'
    DOUBLE = 'double'
    STRING = 'String'
    STRING_1 = 'string'
    STRUCTURE = 'Structure'
    SEQUENCE = 'Sequence'
    OPAQUE = 'opaque'
    ENUM1 = 'enum1'
    ENUM2 = 'enum2'
    ENUM4 = 'enum4'


class ObjectType(Enum):
    ATTRIBUTE = 'attribute'
    DIMENSION = 'dimension'
    VARIABLE = 'variable'
    GROUP = 'group'


@dataclass
class CacheVariable:
    class Meta:
        name = 'cacheVariable'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )


@dataclass
class Dimension:
    class Meta:
        name = 'dimension'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    length: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    is_unlimited: bool = field(
        default=False,
        metadata={
            'name': 'isUnlimited',
            'type': 'Attribute',
        },
    )
    is_variable_length: bool = field(
        default=False,
        metadata={
            'name': 'isVariableLength',
            'type': 'Attribute',
        },
    )
    is_shared: bool = field(
        default=True,
        metadata={
            'name': 'isShared',
            'type': 'Attribute',
        },
    )
    org_name: Optional[str] = field(
        default=None,
        metadata={
            'name': 'orgName',
            'type': 'Attribute',
        },
    )


@dataclass
class LogicalReduce:
    class Meta:
        name = 'logicalReduce'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    dim_names: Optional[str] = field(
        default=None,
        metadata={
            'name': 'dimNames',
            'type': 'Attribute',
            'required': True,
        },
    )


@dataclass
class LogicalSection:
    class Meta:
        name = 'logicalSection'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    section: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )


@dataclass
class LogicalSlice:
    class Meta:
        name = 'logicalSlice'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    dim_name: Optional[str] = field(
        default=None,
        metadata={
            'name': 'dimName',
            'type': 'Attribute',
            'required': True,
        },
    )
    index: Optional[int] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )


@dataclass
class PromoteGlobalAttribute:
    class Meta:
        name = 'promoteGlobalAttribute'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    org_name: Optional[str] = field(
        default=None,
        metadata={
            'name': 'orgName',
            'type': 'Attribute',
        },
    )


@dataclass
class Values:
    class Meta:
        name = 'values'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    start: Optional[float] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    increment: Optional[float] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    npts: Optional[int] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    separator: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    from_attribute: Optional[str] = field(
        default=None,
        metadata={
            'name': 'fromAttribute',
            'type': 'Attribute',
        },
    )
    content: List[object] = field(
        default_factory=list,
        metadata={
            'type': 'Wildcard',
            'namespace': '##any',
            'mixed': True,
        },
    )


@dataclass
class Attribute:
    class Meta:
        name = 'attribute'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    type: DataType = field(
        default=DataType.STRING,
        metadata={
            'type': 'Attribute',
        },
    )
    value: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    separator: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    org_name: Optional[str] = field(
        default=None,
        metadata={
            'name': 'orgName',
            'type': 'Attribute',
        },
    )
    is_unsigned: Optional[bool] = field(
        default=None,
        metadata={
            'name': 'isUnsigned',
            'type': 'Attribute',
        },
    )
    content: List[object] = field(
        default_factory=list,
        metadata={
            'type': 'Wildcard',
            'namespace': '##any',
            'mixed': True,
        },
    )


@dataclass
class EnumTypedef:
    class Meta:
        name = 'enumTypedef'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    enum: List['EnumTypedef.EnumType'] = field(
        default_factory=list,
        metadata={
            'type': 'Element',
            'min_occurs': 1,
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    type: DataType = field(
        default=DataType.ENUM1,
        metadata={
            'type': 'Attribute',
        },
    )

    @dataclass
    class EnumType:
        key: Optional[int] = field(
            default=None,
            metadata={
                'type': 'Attribute',
                'required': True,
            },
        )
        content: List[object] = field(
            default_factory=list,
            metadata={
                'type': 'Wildcard',
                'namespace': '##any',
                'mixed': True,
            },
        )


@dataclass
class Remove:
    class Meta:
        name = 'remove'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    type: Optional[ObjectType] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )


@dataclass
class Variable:
    class Meta:
        name = 'variable'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    attribute: List[Attribute] = field(
        default_factory=list,
        metadata={
            'type': 'Element',
        },
    )
    values: Optional[Values] = field(
        default=None,
        metadata={
            'type': 'Element',
        },
    )
    variable: List['Variable'] = field(
        default_factory=list,
        metadata={
            'type': 'Element',
        },
    )
    logical_section: Optional[LogicalSection] = field(
        default=None,
        metadata={
            'name': 'logicalSection',
            'type': 'Element',
        },
    )
    logical_slice: Optional[LogicalSlice] = field(
        default=None,
        metadata={
            'name': 'logicalSlice',
            'type': 'Element',
        },
    )
    logical_reduce: Optional[LogicalReduce] = field(
        default=None,
        metadata={
            'name': 'logicalReduce',
            'type': 'Element',
        },
    )
    remove: List[Remove] = field(
        default_factory=list,
        metadata={
            'type': 'Element',
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    type: Optional[DataType] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    typedef: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    shape: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    org_name: Optional[str] = field(
        default=None,
        metadata={
            'name': 'orgName',
            'type': 'Attribute',
        },
    )


@dataclass
class Group:
    class Meta:
        name = 'group'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    choice: List[object] = field(
        default_factory=list,
        metadata={
            'type': 'Elements',
            'choices': (
                {
                    'name': 'enumTypedef',
                    'type': EnumTypedef,
                },
                {
                    'name': 'dimension',
                    'type': Dimension,
                },
                {
                    'name': 'variable',
                    'type': Variable,
                },
                {
                    'name': 'attribute',
                    'type': Attribute,
                },
                {
                    'name': 'group',
                    'type': Type['Group'],
                },
                {
                    'name': 'remove',
                    'type': Remove,
                },
            ),
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    org_name: Optional[str] = field(
        default=None,
        metadata={
            'name': 'orgName',
            'type': 'Attribute',
        },
    )


@dataclass
class Aggregation:
    class Meta:
        name = 'aggregation'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    choice: List[object] = field(
        default_factory=list,
        metadata={
            'type': 'Elements',
            'choices': (
                {
                    'name': 'group',
                    'type': Group,
                },
                {
                    'name': 'dimension',
                    'type': Dimension,
                },
                {
                    'name': 'variable',
                    'type': Variable,
                },
                {
                    'name': 'attribute',
                    'type': Attribute,
                },
                {
                    'name': 'remove',
                    'type': Remove,
                },
            ),
        },
    )
    variable_agg: List['Aggregation.VariableAgg'] = field(
        default_factory=list,
        metadata={
            'name': 'variableAgg',
            'type': 'Element',
        },
    )
    promote_global_attribute: List[PromoteGlobalAttribute] = field(
        default_factory=list,
        metadata={
            'name': 'promoteGlobalAttribute',
            'type': 'Element',
        },
    )
    cache_variable: List[CacheVariable] = field(
        default_factory=list,
        metadata={
            'name': 'cacheVariable',
            'type': 'Element',
        },
    )
    netcdf: List['Netcdf'] = field(
        default_factory=list,
        metadata={
            'type': 'Element',
        },
    )
    scan: List['Aggregation.Scan'] = field(
        default_factory=list,
        metadata={
            'type': 'Element',
        },
    )
    scan_fmrc: List['Aggregation.ScanFmrc'] = field(
        default_factory=list,
        metadata={
            'name': 'scanFmrc',
            'type': 'Element',
        },
    )
    type: Optional[AggregationType] = field(
        default=None,
        metadata={
            'type': 'Attribute',
            'required': True,
        },
    )
    dim_name: Optional[str] = field(
        default=None,
        metadata={
            'name': 'dimName',
            'type': 'Attribute',
        },
    )
    recheck_every: Optional[str] = field(
        default=None,
        metadata={
            'name': 'recheckEvery',
            'type': 'Attribute',
        },
    )
    time_units_change: Optional[bool] = field(
        default=None,
        metadata={
            'name': 'timeUnitsChange',
            'type': 'Attribute',
        },
    )
    fmrc_definition: Optional[str] = field(
        default=None,
        metadata={
            'name': 'fmrcDefinition',
            'type': 'Attribute',
        },
    )

    @dataclass
    class VariableAgg:
        name: Optional[str] = field(
            default=None,
            metadata={
                'type': 'Attribute',
                'required': True,
            },
        )

    @dataclass
    class Scan:
        location: Optional[str] = field(
            default=None,
            metadata={
                'type': 'Attribute',
                'required': True,
            },
        )
        reg_exp: Optional[str] = field(
            default=None,
            metadata={
                'name': 'regExp',
                'type': 'Attribute',
            },
        )
        suffix: Optional[str] = field(
            default=None,
            metadata={
                'type': 'Attribute',
            },
        )
        subdirs: bool = field(
            default=True,
            metadata={
                'type': 'Attribute',
            },
        )
        older_than: Optional[str] = field(
            default=None,
            metadata={
                'name': 'olderThan',
                'type': 'Attribute',
            },
        )
        date_format_mark: Optional[str] = field(
            default=None,
            metadata={
                'name': 'dateFormatMark',
                'type': 'Attribute',
            },
        )
        enhance: Optional[bool] = field(
            default=None,
            metadata={
                'type': 'Attribute',
            },
        )

    @dataclass
    class ScanFmrc:
        location: Optional[str] = field(
            default=None,
            metadata={
                'type': 'Attribute',
                'required': True,
            },
        )
        reg_exp: Optional[str] = field(
            default=None,
            metadata={
                'name': 'regExp',
                'type': 'Attribute',
            },
        )
        suffix: Optional[str] = field(
            default=None,
            metadata={
                'type': 'Attribute',
            },
        )
        subdirs: bool = field(
            default=True,
            metadata={
                'type': 'Attribute',
            },
        )
        older_than: Optional[str] = field(
            default=None,
            metadata={
                'name': 'olderThan',
                'type': 'Attribute',
            },
        )
        run_date_matcher: Optional[str] = field(
            default=None,
            metadata={
                'name': 'runDateMatcher',
                'type': 'Attribute',
            },
        )
        forecast_date_matcher: Optional[str] = field(
            default=None,
            metadata={
                'name': 'forecastDateMatcher',
                'type': 'Attribute',
            },
        )
        forecast_offset_matcher: Optional[str] = field(
            default=None,
            metadata={
                'name': 'forecastOffsetMatcher',
                'type': 'Attribute',
            },
        )


@dataclass
class Netcdf:
    class Meta:
        name = 'netcdf'
        namespace = 'http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2'

    read_metadata: Optional[object] = field(
        default=None,
        metadata={
            'name': 'readMetadata',
            'type': 'Element',
        },
    )
    explicit: Optional[object] = field(
        default=None,
        metadata={
            'type': 'Element',
        },
    )
    iosp_param: Optional[object] = field(
        default=None,
        metadata={
            'name': 'iospParam',
            'type': 'Element',
        },
    )
    choice: List[object] = field(
        default_factory=list,
        metadata={
            'type': 'Elements',
            'choices': (
                {
                    'name': 'enumTypedef',
                    'type': EnumTypedef,
                },
                {
                    'name': 'group',
                    'type': Group,
                },
                {
                    'name': 'dimension',
                    'type': Dimension,
                },
                {
                    'name': 'variable',
                    'type': Variable,
                },
                {
                    'name': 'attribute',
                    'type': Attribute,
                },
                {
                    'name': 'remove',
                    'type': Remove,
                },
                {
                    'name': 'aggregation',
                    'type': Aggregation,
                },
            ),
        },
    )
    location: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    title: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    enhance: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    add_records: Optional[bool] = field(
        default=None,
        metadata={
            'name': 'addRecords',
            'type': 'Attribute',
        },
    )
    iosp: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    iosp_param_attribute: Optional[str] = field(
        default=None,
        metadata={
            'name': 'iospParam',
            'type': 'Attribute',
        },
    )
    buffer_size: Optional[int] = field(
        default=None,
        metadata={
            'name': 'bufferSize',
            'type': 'Attribute',
        },
    )
    ncoords: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )
    coord_value: Optional[str] = field(
        default=None,
        metadata={
            'name': 'coordValue',
            'type': 'Attribute',
        },
    )
    section: Optional[str] = field(
        default=None,
        metadata={
            'type': 'Attribute',
        },
    )



#!/usr/bin/env python

"""The setup script."""

import os

from setuptools import find_packages, setup

if os.path.exists('requirements.txt'):
    with open('requirements.txt') as f:
        install_requires = f.read().strip().split('\n')
else:
    install_requires = ['xmltodict', 'xsdata', 'xarray']

if os.path.exists('README.md'):
    with open('README.md') as f:
        long_description = f.read()
else:
    long_description = ''


setup(
    name='xncml',
    description='Tools for manipulating and opening NCML (NetCDF Markup) files with/for xarray',
    long_description=long_description,
    maintainer='Anderson Banihirwe',
    maintainer_email='abanihi@ucar.edu',
    url='https://github.com/xarray-contrib/xncml',
    packages=find_packages(),
    package_dir={'xncml': 'xncml'},
    include_package_data=True,
    install_requires=install_requires,
    license='Apache 2.0',
    zip_safe=False,
    keywords='xncml, xarray, netcdf',
    use_scm_version=True,
    python_requires='>=3.9, <4',
    setup_requires=['setuptools_scm', 'setuptools>=30.3.0', 'setuptools_scm_git_archive'],
    extras_require={
        'dev': ['pytest', 'flake8'],
    },
)
